"""add activated to ReleaseMonitor

Revision ID: bd408a2fd1a5
Revises: 218daaf29e40
Create Date: 2017-11-13 00:31:25.980808

"""

# revision identifiers, used by Alembic.
revision = 'bd408a2fd1a5'
down_revision = '218daaf29e40'

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('release_monitor', sa.Column('activated', sa.Boolean(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    pass
